# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


cimport eckit_geo
from libcpp.utility cimport pair
from libcpp.vector cimport vector

cimport eckit

eckit.eckit_main_initialise()


def version() -> str:
    return eckit_geo.LibEcKitGeo.instance().version()


def git_sha1() -> str:
    return eckit_geo.LibEcKitGeo.instance().gitsha1(40)


def cache_dir_purge() -> None:
    eckit_geo.LibEcKitGeo.purgeCacheDir()


cdef class Area:
    cdef const eckit_geo.Area* _area

    def __dealloc__(self):
        if self._area != NULL:
            del self._area

    def __cinit__(self, spec = None, **kwargs):
        self._area = NULL
        assert bool(spec) != bool(kwargs)

        if kwargs or isinstance(spec, dict):
            from yaml import dump
            spec = dump(kwargs if kwargs else spec, default_flow_style=True).strip()

        try:
            assert isinstance(spec, str)
            self._area = eckit_geo.AreaFactory.make_from_string(spec)

        except RuntimeError as e:
            # opportunity to do something interesting
            raise

    def __eq__(self, other) -> bool:
        if not isinstance(other, Grid):
            return NotImplemented
        return self.spec_str == other.spec_str

    @property
    def spec_str(self) -> str:
        return self._area.spec_str()

    @property
    def spec(self) -> dict:
        from yaml import safe_load
        return safe_load(self.spec_str)

    @property
    def type(self) -> str:
        return self._area.type()


cdef class Grid:
    cdef const eckit_geo.Grid* _grid

    def __dealloc__(self):
        if self._grid != NULL:
            del self._grid

    def __cinit__(self, spec = None, **kwargs):
        self._grid = NULL
        assert bool(spec) != bool(kwargs)

        if kwargs or isinstance(spec, dict):
            from yaml import dump
            spec = dump(kwargs if kwargs else spec, default_flow_style=True).strip()

        try:
            assert isinstance(spec, str)
            self._grid = eckit_geo.GridFactory.make_from_string(spec)

        except RuntimeError as e:
            # opportunity to do something interesting
            raise

    def __eq__(self, other) -> bool:
        if not isinstance(other, Grid):
            return NotImplemented
        return self.spec_str == other.spec_str

    def to_latlons(self):
        cdef pair[vector[double], vector[double]] latlons = self._grid.to_latlons()
        return list(latlons.first), list(latlons.second)

    def distinct_latitudes(self):
        return self._grid.distinct_latitudes()

    def distinct_longitudes(self):
        return self._grid.distinct_longitudes()

    def bounding_box(self) -> tuple:
        cdef const eckit_geo.BoundingBox* bbox = &self._grid.boundingBox()
        cdef double north = bbox.north()
        cdef double west = bbox.west()
        cdef double south = bbox.south()
        cdef double east = bbox.east()
        return north, west, south, east

    def grid_box_areas(self):
        try:
            import mir
        except ModuleNotFoundError:
            return None

        return mir.grid_box_areas(self)

    @property
    def spec_str(self) -> str:
        return self._grid.spec_str()

    @property
    def spec(self) -> dict:
        from yaml import safe_load
        return safe_load(self.spec_str)

    @property
    def catalog_str(self) -> str:
        return self._grid.catalog_str()

    @property
    def catalog(self) -> dict:
        from yaml import safe_load
        return safe_load(self.catalog_str)

    @property
    def type(self) -> str:
        return self._grid.type()

    @property
    def uid(self) -> str:
        return self._grid.uid()

    @property
    def order(self) -> str:
        return self._grid.order()

    @property
    def shape(self) -> tuple:
        cdef vector[size_t] shape_vec = self._grid.shape()
        return tuple(shape_vec)

    def size(self) -> int:
        return self._grid.size()

    def __len__(self) -> int:
        return self.size()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Imports -----------------------------------------------------------------

import sys
import os
import datetime
import re
import subprocess


# -- Path manipulation--------------------------------------------------------

sys.path.append(os.path.abspath("../tests"))


# -- Run Doxygen -------------------------------------------------------------

# Generate Doxygen documentation in the XML format.
subprocess.check_call("doxygen Doxyfile", shell=True)


# -- Project information -----------------------------------------------------

project = "EcKit"
author = "ECMWF"

year = datetime.datetime.now().year
years = "1996-%s" % (year,)

copyright = "%s, %s" % (years, author)


def parse_version(ver_str):
    return re.sub("^((([0-9]+)\\.)+([0-9]+)).*", "\\1", ver_str)


here = os.path.abspath(os.path.dirname(__file__))

# Get the current version directly from the source.
with open(os.path.join(here, "..", "VERSION"), "r") as f:
    release = f.readline().strip()  # full version string

version = parse_version(release)  # feature version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "pydata_sphinx_theme",
    "breathe",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.mermaid",
    "sphinx_copybutton",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

source_suffix = ".rst"
master_doc = "index"
pygments_style = "sphinx"

highlight_language = "c++"


# -- Options for HTML output -------------------------------------------------
html_theme = "pydata_sphinx_theme"
# Do not show source link on each page
html_show_sourcelink = False
html_sidebars = {
    "**": [],
}
html_theme_options = {
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["navbar-icon-links", "theme-switcher", "version-switcher"],
    "navbar_persistent": ["search-button"],
    "primary_sidebar_end": [],
    # On local builds no version.json is present
    "check_switcher": False,
    "logo": {"text": "EcKit"},
}

html_context = {
    # Enable auto detection of light/dark mode
    "default_mode": "auto"
}

# -- Breathe configuration ---------------------------------------------------
breathe_projects = {
    "ECKIT": os.path.abspath("_build/xml")  # Path to generated XML
}
breathe_default_project = "ECKIT"



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


import pytest

from eckit.geo import Grid

SPECS = [
    (dict(grid="1/1"), (181, 360)),
    (dict(grid="H2", ordering="nested"), (12 * 2 * 2,)),
    (dict(grid=[2, 2]), (91, 180)),
    ("{grid: 3/3}", (61, 120)),
    (dict(grid="o2"), (sum([20, 24, 24, 20]),)),
]


@pytest.mark.parametrize("spec, shape", SPECS)
def test_shape(spec, shape):
    grid = Grid(spec)
    assert shape == grid.shape


def test_unstructured_ll():
    grid = Grid(latitudes=[1, 2, 3], longitudes=[4, 5, 6])

    assert grid.spec == dict(type="unstructured_ll", uid=grid.uid)
    assert grid == Grid(uid=grid.uid)

    name = "another-custom-grid"
    grid = Grid(longitudes=[4, 5, 6], latitudes=[1, 2, 3], name=name)

    assert grid == Grid(name)
    assert grid.spec == dict(grid=name)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


import io
import math

import pytest
from eckit.stream import Stream


def _roundtrip(write_fn, read_fn, value):
    """Write value to buffer, read it back, return result."""
    buf = io.BytesIO()
    s = Stream(buf)
    write_fn(s, value)
    buf.seek(0)
    return read_fn(Stream(buf))


@pytest.mark.parametrize("v", [0, 1, 255, 65535, 2**31 - 1])
def test_int_roundtrip(v):
    assert _roundtrip(Stream.write_int, Stream.read_int, v) == v


@pytest.mark.parametrize("v", [0, 1, 100, 2**32 - 1])
def test_unsigned_int_roundtrip(v):
    assert _roundtrip(Stream.write_unsigned_int, Stream.read_unsigned_int, v) == v


@pytest.mark.parametrize("v", [0, 1, 42, 2**32 - 1])
def test_unsigned_long_roundtrip(v):
    assert _roundtrip(Stream.write_unsigned_long, Stream.read_unsigned_long, v) == v


@pytest.mark.parametrize("v", [0, 1, 2**32 - 1])
def test_unsigned_long_long_roundtrip(v):
    # write_unsigned_long_long hard-codes 0 for the high 32 bits — values > 2**32-1 not supported
    assert (
        _roundtrip(Stream.write_unsigned_long_long, Stream.read_unsigned_long_long, v)
        == v
    )


@pytest.mark.parametrize(
    "v",
    [
        0.0,
        1.0,
        -1.0,
        3.14159265358979323846,
        -2.718281828459045,
        1e-300,
        1e300,
        float("inf"),
        float("-inf"),
        1.1754943508222875e-38,  # min normal float32
    ],
)
def test_double_roundtrip(v):
    result = _roundtrip(Stream.write_double, Stream.read_double, v)
    if math.isnan(v):
        assert math.isnan(result)
    elif math.isinf(v):
        assert math.isinf(result) and math.copysign(1, result) == math.copysign(1, v)
    else:
        assert result == v


def test_double_nan_roundtrip():
    nan = float("nan")
    result = _roundtrip(Stream.write_double, Stream.read_double, nan)
    assert math.isnan(result)


@pytest.mark.parametrize("v", [b"", b"hello", b"eckit stream test", b"\x00\x01\x02"])
def test_string_roundtrip(v):
    assert _roundtrip(Stream.write_string, Stream.read_string, v) == v


@pytest.mark.parametrize(
    "data",
    [
        b"",
        b"\x00" * 1,
        bytes(range(256)),
        b"binary\x00data\xff\xfe",
        b"x" * 65536,
    ],
)
def test_large_blob_roundtrip(data):
    assert _roundtrip(Stream.write_large_blob, Stream.read_large_blob, data) == data


def test_multiple_values_sequential():
    buf = io.BytesIO()
    s = Stream(buf)
    s.write_unsigned_long(42)
    s.write_double(3.14)
    s.write_large_blob(b"hello")
    s.write_unsigned_long(99)

    buf.seek(0)
    s2 = Stream(buf)
    assert s2.read_unsigned_long() == 42
    assert s2.read_double() == pytest.approx(3.14)
    assert s2.read_large_blob() == b"hello"
    assert s2.read_unsigned_long() == 99


def test_end_object_skipped_between_reads():
    buf = io.BytesIO()
    s = Stream(buf)
    s.write_unsigned_long(10)
    s.end_object()  # TAG_END_OBJ interleaved
    s.write_unsigned_long(20)

    buf.seek(0)
    s2 = Stream(buf)
    assert s2.read_unsigned_long() == 10
    assert s2.read_unsigned_long() == 20  # TAG_END_OBJ skipped transparently


def test_wrong_tag_raises():
    buf = io.BytesIO()
    s = Stream(buf)
    s.write_unsigned_long(1)

    buf.seek(0)
    with pytest.raises(RuntimeError, match="Unexpected tag"):
        Stream(buf).read_int()  # expects TAG_INT, gets TAG_UNSIGNED_LONG


def test_read_long_long_not_implemented():
    buf = io.BytesIO()
    s = Stream(buf)
    s.write_long_long(1)
    buf.seek(0)
    with pytest.raises(NotImplementedError):
        Stream(buf).read_long_long()



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


def configure_ca_bundle_from_certifi() -> None:
    """Set the CA bundle path for curl from certifi.

    This configures the default CA bundle to all new curl handles.
    """

    try:
        import certifi
    except ImportError:
        return

    ca_bundle = certifi.where()
    if ca_bundle:
        from eckit._eckit import easycurl_setopt

        easycurl_setopt("CURLOPT_CAINFO", ca_bundle)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import findlibs

findlibs.load("eckit_geo", "eckitlib")

from eckit._certifi import configure_ca_bundle_from_certifi
from eckit.geo._eckit_geo import *

configure_ca_bundle_from_certifi()

__lib_version__ = version()
__git_sha1__ = git_sha1()



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


from libcpp.string cimport string as cppstring

cimport eckit

eckit.eckit_main_initialise()


def version() -> str:
    return eckit.LibEcKit.instance().version()


def git_sha1() -> str:
    return eckit.LibEcKit.instance().gitsha1(40)


def easycurl_setopt(str name, object value) -> None:
    """Set a CURL option default that applies to new handles.

    Args:
        name: option (e.g., "CAINFO", "VERBOSE")
        value: bool, int, or str depending on the option

    Examples:
        easycurl_setopt("VERBOSE", True)
        easycurl_setopt("PORT", 8080)
        easycurl_setopt("CAINFO", "/path/to/ca-bundle.crt")
    """
    cdef cppstring c_name = name.encode("utf-8")
    cdef cppstring c_value

    if isinstance(value, bool):
        eckit.easycurl_setopt_bool(c_name, value)
    elif isinstance(value, int):
        eckit.easycurl_setopt_long(c_name, <long>value)
    elif isinstance(value, str):
        c_value = value.encode("utf-8")
        eckit.easycurl_setopt_string(c_name, c_value)
    else:
        raise TypeError(f"Unsupported type for CURL option value: {type(value)}")


def easycurl_clear_options() -> None:
    """Clear all registered CURL options defaults.

    Resets the CURL options defaults that apply to new handles.
    """
    eckit.easycurl_clear_options()



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


from libcpp.string cimport string


cdef extern from "<sstream>" namespace "std" nogil:
    cdef cppclass ostream:
        pass

    cdef cppclass ostringstream(ostream):
        ostringstream()
        string to_string "str" () const



#!/usr/bin/env python3

import os
import signal
from flask import Flask, request, jsonify

"""

MockREST provides a simple REST API over HTTP to support tests.

The API requires a Python environment with `flask` installed to execute -- see requirements.txt.

The API consists of the following end-points:

 * /ping [GET]
    - always replies "Hi!" (200)
 * /insufficient-permissions [GET]
    - always replies "Forbidden" (403)
 * /shutdown [GET]
    - shuts down the running server
 * /blobs [GET]
    - replies with the list of currently available blobs (in JSON)

 * /blob/<key> [GET]
    - replies with the blob associated with <key> (in JSON) (200)
    - if <key> not found, replies with "Not found" (404)
 
 * /blob [POST]
    - creates a new blob, based on payload '{"key": "<key>", "value": "<value>"}'
    - replies with "Request must be JSON" (415) is payload is not JSON
    - replies with "Bad Request" (400) if payload doesn't include "key" or "value"
    - replies with "Bad Request" (400) if a blob is already associated with <key>
    - replies with "Created" (201), in case of success

 * /blob [PUT]
    - updates an existing blob, based on payload '{"key": "<key>", "value": "<value>"}'
    - replies with "Request must be JSON" (415) is payload is not JSON
    - replies with "Bad Request" (400) if payload doesn't include "key" or "value"
    - replies with "Bad Request" (400) if no blob is associated with <key>
    - replies with "OK" (200), in case of success
 
 * /blob/<key> [DELETE]
    - deletes an existing blob, associated with <key>
    - if <key> not found, replies with "Not found" (404)
    - replies with "" (204), in case of success
 
 * /blob/<key>/content [GET]
    - replies with the blob value (in Plain Text) (200)
    - if <key> not found, replies with "Not found" (404)

"""

app = Flask(__name__)


blobs = {}


def _blob_has_all_of(blob, fields):
    for field in fields:
        if field not in blob:
            return False, field
    return True, None


@app.get("/ping")
def ping():
    return "Hi!", 200


@app.get("/insufficient-permissions")
def no_permissions():
    return "Forbidden", 403


@app.get("/shutdown")
def shutdown():
    # terminate own running process!...
    os.kill(os.getpid(), signal.SIGINT)
    return jsonify(blobs)


@app.get("/blobs")
def get_blobs():
    return jsonify(blobs)


@app.get("/blob/<key>")
def get_blob(key):
    if key not in blobs:
        return {"error": f"Not Found: Unable to find blob with key '{key}'"}, 404
    blob = {"key": key, "value": blobs[key]}
    return jsonify(blob)


@app.get("/blob/<key>/content")
def get_blob_content(key):
    if key not in blobs:
        return {"error": f"Not Found: Unable to find blob with key '{key}'"}, 404
    return blobs[key]


@app.post("/blob")
def create_blob():
    if not request.is_json:
        return {"error": "Request must be JSON"}, 415

    blob = request.get_json()

    found, field = _blob_has_all_of(blob, ["key", "value"])
    if not found:
        return {"error": f"Bad Request: New blob must contain field '{field}'"}, 400

    key = blob["key"]
    if key in blobs:
        return {"error": f"Bad Request: Key {blob['key']} already exists"}, 400

    value = blob["value"]
    blobs[key] = value
    return f"Created: {blob}", 201


@app.put("/blob")
def update_blob():
    if not request.is_json:
        return {"error": "Request must be JSON"}, 415

    blob = request.get_json()

    found, field = _blob_has_all_of(blob, ["key", "value"])
    if not found:
        return {"error": f"Bad Request: Update blob must contain field '{field}'"}, 400

    key = blob["key"]
    if key not in blobs:
        return {"error": f"Bad Request: Key {blob['key']} does not exist"}, 400

    value = blob["value"]
    blobs[key] = value
    return f"OK: {blob}", 200


@app.delete("/blob/<key>")
def delete_blob(key):
    if key not in blobs:
        return {"error": f"Not Found (Unable to find blob with key '{key}')"}, 404
    del blobs[key]
    return "", 204



import platform
import sys

from setuptools import setup
from wheel.bdist_wheel import bdist_wheel

with open("VERSION", "r") as fVersion:
    version = fVersion.readlines()[0].strip()
install_requires = [
    f"eckitlib=={version}",
    "findlibs",
    "pyyaml",
    "requests",
]


# NOTE see ci-utils/wheelmaker/buildscripts/setup_utils, we need to get the right abi compat tag
class bdist_wheel_ext(bdist_wheel):
    def get_tag(self):
        python, abi, plat = bdist_wheel.get_tag(self)  # noqa: F841
        return python, abi, f"manylinux_2_28_{platform.machine()}"


ext_kwargs = {
    "darwin": {},
    "linux": {"cmdclass": {"bdist_wheel": bdist_wheel_ext}},
}

setup(
    version=version,
    packages=["eckit"],
    package_data={
        "": ["*.so"],
    },
    has_ext_modules=lambda: True,
    install_requires=install_requires,
    **ext_kwargs[sys.platform],
)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


from libcpp.string cimport string
from libcpp.utility cimport pair
from libcpp.vector cimport vector


cdef extern from "eckit/geo/LibEcKitGeo.h" namespace "eckit":
    cdef cppclass LibEcKitGeo:
        @staticmethod
        LibEcKitGeo& instance()

        @staticmethod
        void purgeCacheDir()

        string version()
        string gitsha1(unsigned int n)  # n=40 for full sha1


cdef extern from "eckit/geo/Area.h" namespace "eckit::geo":
    cdef cppclass Area:
        string spec_str() const
        string type() const

    cdef cppclass AreaFactory:
        @staticmethod
        const Area* make_from_string(const string) except +


cdef extern from "eckit/geo/area/BoundingBox.h" namespace "eckit::geo::area":
    cdef cppclass BoundingBox(Area):
        double north() const
        double west() const
        double south() const
        double east() const


cdef extern from "eckit/geo/Grid.h" namespace "eckit::geo":
    cdef cppclass Grid:
        string spec_str() const
        string catalog_str() const
        string type() const
        string uid() const
        string order() const

        pair[vector[double], vector[double]] to_latlons() const
        vector[double] distinct_latitudes() const
        vector[double] distinct_longitudes() const

        vector[size_t] shape() const
        size_t size() const
        const BoundingBox& boundingBox() const

    cdef cppclass GridFactory:
        @staticmethod
        const Grid* make_from_string(const string) except +



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

from struct import pack
from struct import unpack

TAG_ZERO = 0
TAG_START_OBJ = 1
TAG_END_OBJ = 2
TAG_CHAR = 3
TAG_UNSIGNED_CHAR = 4
TAG_INT = 5
TAG_UNSIGNED_INT = 6
TAG_SHORT = 7
TAG_UNSIGNED_SHORT = 8
TAG_LONG = 9
TAG_UNSIGNED_LONG = 10
TAG_LONG_LONG = 11
TAG_UNSIGNED_LONG_LONG = 12
TAG_FLOAT = 13
TAG_DOUBLE = 14
TAG_STRING = 15
TAG_BLOB = 16
TAG_EXCEPTION = 17
TAG_START_REC = 18
TAG_END_REC = 19
TAG_EOF = 20
TAG_LARGE_BLOB = 21  # For blobs >= 2Gb
LAST_TAG = 22

TAG_NAME = (
    "0",
    "start of object",
    "end of object",
    "char",
    "unsigned char",
    "int",
    "unsigned int",
    "short",
    "unsigned short",
    "long",
    "unsigned long",
    "long long",
    "unsigned long long",
    "float",
    "double",
    "string",
    "blob",
    "exception",
    "start of record",
    "end of record",
)


class Stream:
    def __init__(self, stream):
        self.stream = stream

    def _read(self, n):
        return self.stream.read(n)

    def _write(self, n, nbytes=None):
        assert not nbytes or nbytes == len(n)
        return self.stream.write(n)

    def read_tag(self, expected_tag):
        tag = self._read(1)
        tag = ord(tag)

        while tag == TAG_END_OBJ:
            tag = self._read(1)
            if tag:
                tag = ord(tag)

        # TODO There is some suspicious code on the perl side for the following.
        # TODO Maybe worth having a look at it sometime.
        if tag != expected_tag:
            if tag < len(TAG_NAME):
                raise RuntimeError(
                    "Unexpected tag: %s, wanted %s"
                    % (TAG_NAME[tag], TAG_NAME[expected_tag])
                )
            else:
                raise RuntimeError("Invalid tag with id: '%s'" % tag)

        return tag

    def write_tag(self, tag):
        self._write(pack("b", tag), 1)

    def next_object(self):
        while 1:
            tag = self.stream.recv(1)
            if tag == "":
                return 0

            tag = unpack("b", tag)[0]

            if tag == TAG_START_OBJ:
                return 1
            else:
                if tag < len(TAG_NAME):
                    raise RuntimeError(
                        "Unexpected tag: '%s', wanted '%s'"
                        % (TAG_NAME[tag], TAG_NAME[TAG_START_OBJ])
                    )
                else:
                    raise RuntimeError("Invalid tag with id: '%d'" % tag)

    def write_char(self, c):
        self.write_tag(TAG_CHAR)
        self._write(c, 1)

    def write_unsigned_char(self, c):
        self.write_tag(TAG_UNSIGNED_CHAR)
        self._write(c, 1)

    def write_int(self, n):
        self.write_tag(TAG_INT)
        self._write(pack("!L", n), 4)

    def write_unsigned_int(self, n):
        self.write_tag(TAG_UNSIGNED_INT)
        self._write(pack("!L", n), 4)

    def write_long(self, n):
        self.write_tag(TAG_LONG)
        self._write(pack("!L", n), 4)

    def write_double(self, n):
        self.write_tag(TAG_DOUBLE)
        x, y = unpack("!II", pack("!d", n))
        self.stream.write(pack("!L", y))
        self.stream.write(pack("!L", x))

    def write_unsigned_long(self, n):
        self.write_tag(TAG_UNSIGNED_LONG)
        self._write(pack("!L", n), 4)

    def write_long_long(self, n):
        self.write_tag(TAG_LONG_LONG)
        self._write(pack("!L", n), 4)

    def write_unsigned_long_long(self, n):
        self.write_tag(TAG_UNSIGNED_LONG_LONG)
        self._write(pack("!L", 0), 4)
        self._write(pack("!L", n), 4)

    def write_string(self, s):
        self.write_tag(TAG_STRING)
        n = len(s)
        self._write(pack("!L", n), 4)
        self._write(s, n)

    def write_large_blob(self, data: bytes):
        self.write_tag(TAG_LARGE_BLOB)
        self._write(pack("!Q", len(data)), 8)
        self._write(data)

    def read_char(self):
        self.read_tag(TAG_CHAR)
        return self._read(1)

    def read_unsigned_char(self):
        self.read_tag(TAG_UNSIGNED_CHAR)
        return self._read(1)

    def read_int(self):
        self.read_tag(TAG_INT)
        return unpack("!L", self._read(4))[0]

    def read_unsigned_int(self):
        self.read_tag(TAG_UNSIGNED_INT)
        return unpack("!L", self._read(4))[0]

    def read_long(self):
        self.read_tag(TAG_LONG)
        return unpack("!L", self._read(4))[0]

    def read_double(self):
        self.read_tag(TAG_DOUBLE)
        y = unpack("!L", self._read(4))[0]  # low word written first by write_double
        x = unpack("!L", self._read(4))[0]  # high word second
        return unpack("!d", pack("!II", x, y))[0]

    def read_unsigned_long(self):
        self.read_tag(TAG_UNSIGNED_LONG)
        return unpack("!L", self._read(4))[0]

    def read_long_long(self):
        self.read_tag(TAG_LONG_LONG)
        raise NotImplementedError(
            "read_long_long: write_long_long only writes 4 bytes, format inconsistent"
        )

    def _read_long_long(self):
        return unpack("!Q", self._read(8))[0]

    def read_unsigned_long_long(self):
        self.read_tag(TAG_UNSIGNED_LONG_LONG)
        return unpack("!Q", self._read(8))[0]

    def read_string(self):
        self.read_tag(TAG_STRING)
        n = unpack("!L", self._read(4))[0]
        return self._read(n)

    def read_large_blob(self):
        self.read_tag(TAG_LARGE_BLOB)
        length = unpack("!Q", self._read(8))[0]
        return self._read(length)

    def start_object(self):
        self.write_tag(TAG_START_OBJ)

    def end_object(self):
        self.write_tag(TAG_END_OBJ)

    def read_object(self):
        if not self.next_object():
            return None
        klass = self.read_string()
        return klass

    def write_object(self, name, ref):
        self.start_object()
        self.write_string(name)
        self.write_string(ref)
        self.end_object()



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


import pytest

SPECS = [
    (dict(north=90, west=0, south=-90, east=360), dict(area=[90, 0, -90, 360])),
    (
        dict(north=10, west=360 * 99 + 1, south=0, east=360 * 42 + 10),
        dict(area=[10, 1, 0, 10]),
    ),
]


@pytest.mark.parametrize("_spec, _expected", SPECS)
def test_area(_spec, _expected):
    from eckit.geo import Area

    area = Area(_spec)
    assert area.spec == _expected



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


cimport std
from libcpp.string cimport string


cdef extern from "eckit.h":
    void eckit_main_initialise()


cdef extern from "eckit/config/LibEcKit.h" namespace "eckit":
    cdef cppclass LibEcKit:
        @staticmethod
        LibEcKit& instance()

        string version()
        string gitsha1(unsigned int n)  # n=40 for full sha1


cdef extern from "eckit/filesystem/PathName.h" namespace "eckit":
    cdef cppclass PathName:
        PathName(string)


cdef extern from "eckit/log/JSON.h" namespace "eckit":
    cdef cppclass JSON:
        JSON(std.ostream&)


cdef extern from "eckit/io/EasyCURL.h" namespace "eckit":
    void easycurl_setopt_bool(string option_name, bool value)
    void easycurl_setopt_long(string option_name, long value)
    void easycurl_setopt_string(string option_name, string value)
    void easycurl_clear_options()
